#ifndef com_sleepless_net_blockserver_cpp
#define com_sleepless_net_blockserver_cpp

// Copyright 2004 Sleepless Software Inc.  All Rights Reserved



#include "log.cpp"
#include "thread.cpp"
#include "system.cpp"
#include "queue.cpp"
#include "serversocket.cpp"
#include "blocksocket.cpp"



// You can override the defaults here by using #define before #including
#ifndef BS_NUM_RUNNERS
#	define BS_NUM_RUNNERS	2
#endif

#ifndef BS_INSZ
#	define BS_INSZ			4000
#endif

#ifndef BS_OUTSZ
#	define BS_OUTSZ			4000
#endif

#ifndef BS_BACKLOG
#	define BS_BACKLOG		5
#endif



struct BlockRunner : Runnable
{
	char buf_in[BS_INSZ];
	int len_in;
	char buf_out[BS_OUTSZ];
	int len_out;
	int done;
	BlockSocket *bs;


	void run()
	{
		unsigned int ep;

		len_out = 0;
		len_in = bs->read(buf_in, BS_INSZ, &ep);
		if(len_in != -1)
		{
			transact();
			bs->write(buf_out, len_out);
		}
		else
		{
			if(ep == 0x47455420)
			{
				// "GET " as a 4-byte integer
				int r;
				len_in = 0;
				int left = BS_INSZ;
				char *bp;
				while(true)
				{
					bp = buf_in + len_in;
					r = bs->sock->readLine(bp, left);
					if(r == -1)
						break;
					Str::trim(bp);
					if(bp[0] == 0)
						break;
					Str::append(bp, "\n", strlen(bp) + 2);
					r = strlen(bp);
					len_in += r;
					left -= r;
				}
				if(r != -1)
				{
					transact_http_get();
					bs->sock->write(buf_out, len_out);
				}
			}
		}
		delete bs;		// dispose of socket
		bs = 0;
		done = 1;
	}


	/* incoming msg has been read into buf_in, and the size of that
	msg is in len_in.
	Any output should be placed into buf_out, and len_out should be
	set to the size of that data. */
	virtual void transact() = 0;


	virtual void transact_http_get() = 0;

};



struct BlockServer
{
	int port;
	Queue r_ready;
	Queue r_running;
	int done;
	int hits;


	virtual ~BlockServer()
	{
	}


	// Move any finished runners on the r_running queue to the ready queue
	void finish_runners()
	{
		int l = r_running.length();
		for(int i = 0; i < l; i++)
		{
			BlockRunner *rr = (BlockRunner *)r_running.next();
			if(rr)
			{
				if(rr->done)
				{
					rr->join();			// finish thread
					rr->done = 0;		// make sure done is cleared
					//delete rr->bs;		// dispose of socket
					r_ready.queue(rr);	// put it in ready queue
				}
			}
		}
	}


	// Returns the next available runner
	BlockRunner *getRunner()
	{
		BlockRunner *rr;

		while(!done)
		{

			finish_runners();

			if(r_ready.length() >= 1)
			{
				rr = (BlockRunner *)r_ready.next();	// take from ready queue
				r_running.queue(rr);			// put on running queue
				break;							// return pointer
			}

			if((r_ready.length() + r_running.length()) < BS_NUM_RUNNERS)
			{
				// make a new runner
				rr = create_runner();
				if(rr)
				{
					r_running.queue(rr);			// put on running queue
					break;
				}
			}

			System::msleep(250);
		}

		return rr;
	}


	void run(int p)
	{
		hits = 0;
		done = 0;
		port = p;
		ServerSocket ss(port, 10);
		if(ss.isListening())
		{
TRACE("Listening on port %d\n", port);
			while(!done)
			{
				Socket *sock = ss.accept();
				hits++;
TRACE("New connection %d\n", hits);

				BlockSocket *bs = new BlockSocket(sock);

				BlockRunner *rr = getRunner();
				rr->bs = bs;
				rr->start();
			}

TRACE("Exiting ...\n");
			for(int i = 0; i < (4 * 5); i++)
			{
				int n = r_running.length();
				if(n == 0)
					break;
TRACE("   waiting for %d threads ...\n", n);
				finish_runners();
				System::msleep(250);
			}
		}
	}


	
	// This should return a newly created subclass of BlockRunner or
	// NULL if one couldn't be created.
	virtual BlockRunner *create_runner() = 0;


};




#ifdef BLOCKSERVER_TEST

int rcount = 0;
int tcount = 0;

Mutex mutex;

struct MyRunner : BlockRunner
{
	int id;


	MyRunner(int n)
	{
		id = n;
	}

	
	void transact()
	{
		mutex.lock();

		buf_in[len_in] = 0;
		sprintf(buf_out,
			"Hello, I am server thread %d of %d.\n"
			"I've received your message '%s'.\n"
			"This is transaction %d.\n"
			"",
			id, rcount,
			buf_in,
			tcount
			);
		len_out = strlen(buf_out);

		mutex.unlock();
	}

	void transact_http_get()
	{
		TRACE("transact_http_get: len_in=%d buf_in='%s'\n", len_in, buf_in);

		sprintf(buf_out,
			"HTTP 200 OK\r\n"
			"Content-Type: text/html\r\n"
			"\r\n"
			"<html><body>Hi <b>Mom</b>.\r\n"
			);
		len_out = strlen(buf_out);
	}

};


struct MyServer : BlockServer
{

	BlockRunner *create_runner()
	{
		return new MyRunner(rcount++);
	}

};




int main(int argc, char **argv)
{
	TRACEON();

	if(argc == 2)
	{
		int port = atoi(argv[1]);
		TRACE("Acting as server: port=%d\n", port);

		MyServer bs;
		bs.run(port);
	}
	else
	if(argc == 4)
	{
		const char *host = argv[1];
		int port = atoi(argv[2]);
		TRACE("Acting as client: host=%s port=%d\n", host, port);

		BlockSocket bs(host, port);
		if(bs.isConnected())
		{
			TRACE("connected\n");
			char s[256];
			strcpy(s, argv[3]);
			int w = bs.write(s, strlen(s));
			if(w != -1)
			{
				TRACE("wrote %d bytes\n", w);
				int r = bs.read(s, sizeof(s));
				if(r != -1)
				{
					TRACE("read %d bytes\n", r);
					s[r] = 0;
					TRACE("response: '%s'\n", s);
				}
				else
					TRACE("Read error\n");
			}
			else
				TRACE("Write error\n");
		}
		else
			TRACE("Couldn't connect\n");

	}

	TRACE("Done.\n");

	return 0;
}


#endif





#endif // com_sleepless_net_blockserver_cpp
