#ifndef com_sleepless_list_cpp
#define com_sleepless_list_cpp

/* Copyright 2004-2304 Sleepless Software Inc.  All Rights Reserved */

/*
	A generic list for holding 'pointers', like pointers to strings, or
	pointers to objects, etc.  Find and remove items using the pointers
	you used to add them.

	Note: This code is not particularly efficient and does a lot
	of traversing of the array, but it's simple and straightforward.

	Note: Because this object uses an Array object, removing an item
	from the list won't actually reduce its memory use.
	See array.cpp 

*/

#include <stdlib.h>

#include "array.cpp"
#include "assert.cpp"


struct List
{
	Array items;
	int size;

	List()
	{
		size = 0;
	}

	// add a pointer to the end of the list
	void append(void *o)
	{
		items.put(size++, o);
	}

	// wedge a pointer into the list at a specific index.
	void insert(void *o, int pos)
	{
		if(pos < 0)
			pos = 0;
		if(pos > size)		// this is supposed to be size, not size-1
			pos = size;
		for(int i = size; i > pos; i--)		// scoot existing items up
			items.put(i, items.get(i - 1));
		size++;
		items.put(pos, o);			// put new one at given index
	}

	/* return pointer (without removing it) at specific index or null if index is out of range */
	void *get(int pos)
	{
		if(size > 0)
			return items.get(pos);		// Array does range check
		return 0;
	}

	/* remove and return pointer at specific index, or null if index out of
		range */
	void *rem(int pos)
	{
		if((pos < 0) || (pos >= size))
				return 0;
		void *o = items.get(pos);	// Array does range check
		for(int i = pos; i < (size - 1); i++)		// scoot existing items up
			items.put(i, items.get(i + 1));
		size--;
		return o;
	}

	// find pointer in list and return it's index, or -1 if not found
	int locate(void *o)
	{
		for(int pos = 0; pos < size; pos++)
		{
			if(items.get(pos) == o)
				return pos;
		}
		return -1;
	}

	// remove and return a pointer from the list, or null if not on the list
	void *rem(void *o)
	{
		int n = locate(o);
		return (n == -1) ? 0 : rem(n);
	}

	// return the number of pointers in the list.
	int length()
	{
		return size;
	}

	// clear the list of all pointers
	void remAll()
	{
		size = 0;
	}


	// XXX need qsort for palmos in order to enable this.
	// XXX See below j.h.
	void sort(int (*cmp)(const void *elem1, const void *elem2))
	{

#ifndef __palmos__
		int l = length();

		const void **ptrs = (const void **)malloc(sizeof(const void *) * l);
		if(!ptrs)
			return;

		int i;

		for(i = 0; i < l; i++)
		{
			ptrs[i] = rem(0);
		}

		qsort(ptrs, l, sizeof(const void *), cmp);

		for(i = 0; i < l; i++)
		{
			append((void *)ptrs[i]);
		}

		free(ptrs);
#endif

	}

#if 0
/*
 * File: qsort.c
 * Author: James Hall (jch1003@cl.cam.ac.uk)
 * Copyright (C) University of Cambridge Computer Laboratory, 1994
 **~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 ** PACKAGE: Nemesis C library.
 **
 ** FUNCTION: Using quicksort order array pbase[0]...pbase[total_elems -1]
 **           of objects size size using comparison function cmp.
 **           RAISES Heap_NoMemory. 
 **
 ** HISTORY:
 ** Created: Wed May 18 11:34:03 1994 (jch1003)
 ** Last Edited: Thu Oct 20 11:59:14 1994 By James Hall
 **
 ** $Id: list.cpp,v 1.6 2004/11/10 03:40:25 joe Exp $
 **
 **~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 */
/* Copyright (C) 1991, 1992 Free Software Foundation, Inc.
This file is part of the GNU C Library.

The GNU C Library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public License as
published by the Free Software Foundation; either version 2 of the
License, or (at your option) any later version.

The GNU C Library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with the GNU C Library; see the file COPYING.LIB.  If
not, write to the Free Software Foundation, Inc., 675 Mass Ave,
Cambridge, MA 02139, USA.  */


//#include <nemesis.h>
//#include <stdlib.h>
//#include <string.h>

//#define TRC(x) 

//#define write printf

/* Byte-wise swap two items of size SIZE. */
#define Q_SWAP(a, b, size)                                                      \
  do                                                                          \
    {                                                                         \
      register size_t __size = (size);                                        \
      register char *__a = (a), *__b = (b);                                   \
      do                                                                      \
        {                                                                     \
          char __tmp = *__a;                                                  \
          *__a++ = *__b;                                                      \
          *__b++ = __tmp;                                                     \
        } while (--__size > 0);                                               \
    } while (0)

/* Discontinue quicksort algorithm when partition gets below this size.
   This particular magic number was chosen to work best on a Sun 4/260. */
#define Q_MAX_THRESH 4

/* Stack node declarations used to store unfulfilled partition obligations. */
typedef struct 
  {
    char *lo;
    char *hi;
  } qsort_stack_node;

/* The next 4 #defines implement a very fast in-line stack abstraction. */
#define Q_STACK_SIZE      (8 * sizeof(unsigned long int))
#define Q_PUSH(low, high) ((void) ((top->lo = (low)), (top->hi = (high)), ++top))
#define Q_POP(low, high)  ((void) (--top, (low = top->lo), (high = top->hi)))
#define Q_STACK_NOT_EMPTY (stack < top)                


/* Order size using quicksort.  This implementation incorporates
   four optimizations discussed in Sedgewick:

   1. Non-recursive, using an explicit stack of pointer that store the 
      next array partition to sort.  To save time, this maximum amount 
      of space required to store an array of MAX_INT is allocated on the 
      stack.  Assuming a 32-bit integer, this needs only 32 * 
      sizeof(qsort_stack_node) == 136 bits.  Pretty cheap, actually.

   2. Chose the pivot element using a median-of-three decision tree.
      This reduces the probability of selecting a bad pivot value and 
      eliminates certain extraneous comparisons.

   3. Only quicksorts TOTAL_ELEMS / Q_MAX_THRESH partitions, leaving
      insertion sort to order the Q_MAX_THRESH items within each partition.  
      This is a big win, since insertion sort is faster for small, mostly
      sorted array segements.

   4. The larger of the two sub-partitions is always pushed onto the
      stack first, with the algorithm then concentrating on the
      smaller partition.  This *guarantees* no more than log (n)
      stack size is needed (actually O(1) in this case)!  */

void qsort(void *pbase, size_t total_elems, size_t size, int(*cmp) (const void *keyval, const void *datum))
{
  register char *base_ptr = (char *) pbase;
  char *pivot_buffer;
  const size_t max_thresh = Q_MAX_THRESH * size;

  /* Allocating SIZE bytes for a pivot buffer facilitates a better
     algorithm below since we can do comparisons directly on the pivot. */

  pivot_buffer = (char *) Heap$Malloc (Pvs(heap), size);

  if (total_elems == 0) {
    /* Avoid lossage with unsigned arithmetic below.  */
    FREE (pivot_buffer);
    return;
  }

  if (total_elems > Q_MAX_THRESH)
    {
      char *lo = base_ptr;
      char *hi = &lo[size * (total_elems - 1)];
      /* Largest size needed for 32-bit int!!! */
      qsort_stack_node stack[Q_STACK_SIZE];
      qsort_stack_node *top = stack + 1;

      while (Q_STACK_NOT_EMPTY)
        {
          char *left_ptr;
          char *right_ptr;

          char *pivot = pivot_buffer;

          /* Select median value from among LO, MID, and HI. Rearrange
             LO and HI so the three values are sorted. This lowers the 
             probability of picking a pathological pivot value and 
             skips a comparison for both the LEFT_PTR and RIGHT_PTR. */

          char *mid = lo + size * ((hi - lo) / size >> 1);

          if ((*cmp)((void*) mid, (void*) lo) < 0)
            Q_SWAP(mid, lo, size);
          if ((*cmp)((void*) hi, (void*) mid) < 0)
            Q_SWAP(mid, hi, size);
          else 
            goto jump_over;
          if ((*cmp)((void*) mid, (void*) lo) < 0)
            Q_SWAP(mid, lo, size);
        jump_over:;
          memcpy(pivot, mid, size);
          pivot = pivot_buffer;

          left_ptr  = lo + size;
          right_ptr = hi - size; 

          /* Here's the famous ``collapse the walls'' section of quicksort.  
             Gotta like those tight inner loops!  They are the main reason 
             that this algorithm runs much faster than others. */
          do 
            {
              while ((*cmp)((void*) left_ptr, (void*) pivot) < 0)
                left_ptr += size;

              while ((*cmp)((void*) pivot, (void*) right_ptr) < 0)
                right_ptr -= size;

              if (left_ptr < right_ptr) 
                {
                  Q_SWAP(left_ptr, right_ptr, size);
                  left_ptr += size;
                  right_ptr -= size;
                }
              else if (left_ptr == right_ptr) 
                {
                  left_ptr += size;
                  right_ptr -= size;
                  break;
                }
            } 
          while (left_ptr <= right_ptr);

          /* Set up pointers for next iteration.  First determine whether
             left and right partitions are below the threshold size.  If so, 
             ignore one or both.  Otherwise, push the larger partition's
             bounds on the stack and continue sorting the smaller one. */

          if ((size_t) (right_ptr - lo) <= max_thresh)
            {
              if ((size_t) (hi - left_ptr) <= max_thresh)
                /* Ignore both small partitions. */
                Q_POP(lo, hi); 
              else
                /* Ignore small left partition. */  
                lo = left_ptr;
            }
          else if ((size_t) (hi - left_ptr) <= max_thresh)
            /* Ignore small right partition. */
            hi = right_ptr;
          else if ((right_ptr - lo) > (hi - left_ptr))
            {                   
              /* Push larger left partition indices. */
              Q_PUSH(lo, right_ptr);
              lo = left_ptr;
            }
          else
            {                   
              /* Push larger right partition indices. */
              Q_PUSH(left_ptr, hi);
              hi = right_ptr;
            }
        }
    }

  /* Once the BASE_PTR array is partially sorted by quicksort the rest
     is completely sorted using insertion sort, since this is efficient 
     for partitions below Q_MAX_THRESH size. BASE_PTR points to the beginning 
     of the array to sort, and END_PTR points at the very last element in
     the array (*not* one beyond it!). */

  /* MIN is #defined in nemesis.h. */

  {
    char *const end_ptr = &base_ptr[size * (total_elems - 1)];
    char *tmp_ptr = base_ptr;
    char *thresh = MIN(end_ptr, base_ptr + max_thresh);
    register char *run_ptr;

    /* Find smallest element in first threshold and place it at the
       array's beginning.  This is the smallest array element,
       and the operation speeds up insertion sort's inner loop. */

    for (run_ptr = tmp_ptr + size; run_ptr <= thresh; run_ptr += size)
      if ((*cmp)((void*) run_ptr, (void*) tmp_ptr) < 0)
        tmp_ptr = run_ptr;

    if (tmp_ptr != base_ptr)
      Q_SWAP(tmp_ptr, base_ptr, size);

    /* Insertion sort, running from left-hand-side up to right-hand-side.  */

    run_ptr = base_ptr + size;
    while ((run_ptr += size) <= end_ptr)
      {
        tmp_ptr = run_ptr - size;
        while ((*cmp)((void*) run_ptr, (void*) tmp_ptr) < 0)
          tmp_ptr -= size;

        tmp_ptr += size;
        if (tmp_ptr != run_ptr)
          {
            char *trav;

            trav = run_ptr + size;
            while (--trav >= run_ptr)
              {
                char c = *trav;
                char *hi, *lo;

                for (hi = lo = trav; (lo -= size) >= tmp_ptr; hi = lo)
                  *hi = *lo;
                *hi = c;
              }
          }
      }
  }

  FREE (pivot_buffer);

}

#endif

};


#if 0

	/* DEFUNCT: This is brutal on a stack, and probably won't work good for
		for lesser platforms such as palm, etc. */
	
struct Node
{
	Node *next;
	const void *object;

	Node(const void *o)
	{
		object = o;
		next = 0;
	}

	void append(Node *n)
	{
		if(next)
			next->append(n);
		else
			next = n;
	}

	// Insert node into list at index position 'pos'.
	// Returns 'n' if 'n' has become the new head of the list.
	// Otherwise, returns 'h'.
	Node *insert(Node *n, int pos, Node *h)
	{
		if(pos <= 0)
		{
			n->next = this;
			return n;
		}
		if(pos == 1)
		{
			n->next = next;
			next = n;
			return h;
		}
		return next->insert(n, pos - 1, h);
	}

	Node *nth(int i)
	{
		if(i == 0)
			return this;
		if(next)
			return next->nth(i - 1);
		else
			return 0;
	}

	int length(int n)
	{
		if(next)
			return next->length(n + 1);
		return n;
	}

	Node *rem(int n, Node *prev)
	{
		if(n < 0)
			return 0;
		if(n == 0)
		{
			if(prev)
				prev->next = next;
			return this;
		}
		if(next)
			return next->rem(n - 1, this);
		return 0;
	}

	int locate(int i, const void *o)
	{
		if(object == o)
			return i;
		if(next)
			return next->locate(i + 1, o);
		return -1;
	}

};


struct List
{
	Node *first;

	List()
	{
		first = 0;
	}


	// add a pointer to end of list
	void append(const void *o)
	{
		Node *node = new Node(o);
		if(first)
			first->append(node);
		else
			first = node;
	}

	// place a pointer in the list at specific index.
	void insert(const void *o, int pos)
	{
		Node *node = new Node(o);
		if(first == 0)
			first = node;
		else
			first = first->insert(node, pos, first);
	}

	// return pointer at specific index
	const void *nth(int n)
	{
		if(first)
		{
			Node *node = first->nth(n);
			if(node)
				return node->object;
		}
		return 0;
	}

	// remove pointer at specific index
	const void *rem(int n)
	{
		if(first)
		{
			Node *node = first->rem(n, 0);
			if(node)
			{
				const void *o = node->object;
				if(node == first)
					first = node->next;
				delete node;
				return o;
			}
		}
		return 0;
	}

	// find pointer in list and return it's index, or -1 if not found
	int locate(const void *o)
	{
		if(first)
			return first->locate(0, o);
		return -1;
	}

	// remove a pointer from the list
	const void *rem(const void *o)
	{
		int n = locate(o);
		return (n == -1) ? 0 : rem(n);
	}

	// return the number of pointers in the list.
	int length()
	{
		if(first)
			return first->length(1);
		return 0;
	}

	// clear the list of all pointers
	void remAll()
	{
		// XXX this is broken ... if a null pointer is stored
		// on the list, this loop will exit prematurely.
		while(rem(0))
			;
	}


	// XXX need qsort for palmos in order to enable this.
	// XXX See below j.h.
	void sort(int (*cmp)(const void *elem1, const void *elem2))
	{

#ifndef __palmos__
		int l = length();

		const void **ptrs = (const void **)malloc(sizeof(const void *) * l);
		if(!ptrs)
			return;

		int i;

		for(i = 0; i < l; i++)
		{
			ptrs[i] = rem(0);
		}

		qsort(ptrs, l, sizeof(const void *), cmp);

		for(i = 0; i < l; i++)
		{
			append(ptrs[i]);
		}

		free(ptrs);
#endif

	}
#if 0
/*
 * File: qsort.c
 * Author: James Hall (jch1003@cl.cam.ac.uk)
 * Copyright (C) University of Cambridge Computer Laboratory, 1994
 **~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 ** PACKAGE: Nemesis C library.
 **
 ** FUNCTION: Using quicksort order array pbase[0]...pbase[total_elems -1]
 **           of objects size size using comparison function cmp.
 **           RAISES Heap_NoMemory. 
 **
 ** HISTORY:
 ** Created: Wed May 18 11:34:03 1994 (jch1003)
 ** Last Edited: Thu Oct 20 11:59:14 1994 By James Hall
 **
 ** $Id: list.cpp,v 1.6 2004/11/10 03:40:25 joe Exp $
 **
 **~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 */
/* Copyright (C) 1991, 1992 Free Software Foundation, Inc.
This file is part of the GNU C Library.

The GNU C Library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public License as
published by the Free Software Foundation; either version 2 of the
License, or (at your option) any later version.

The GNU C Library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with the GNU C Library; see the file COPYING.LIB.  If
not, write to the Free Software Foundation, Inc., 675 Mass Ave,
Cambridge, MA 02139, USA.  */


//#include <nemesis.h>
//#include <stdlib.h>
//#include <string.h>

//#define TRC(x) 

//#define write printf

/* Byte-wise swap two items of size SIZE. */
#define Q_SWAP(a, b, size)                                                      \
  do                                                                          \
    {                                                                         \
      register size_t __size = (size);                                        \
      register char *__a = (a), *__b = (b);                                   \
      do                                                                      \
        {                                                                     \
          char __tmp = *__a;                                                  \
          *__a++ = *__b;                                                      \
          *__b++ = __tmp;                                                     \
        } while (--__size > 0);                                               \
    } while (0)

/* Discontinue quicksort algorithm when partition gets below this size.
   This particular magic number was chosen to work best on a Sun 4/260. */
#define Q_MAX_THRESH 4

/* Stack node declarations used to store unfulfilled partition obligations. */
typedef struct 
  {
    char *lo;
    char *hi;
  } qsort_stack_node;

/* The next 4 #defines implement a very fast in-line stack abstraction. */
#define Q_STACK_SIZE      (8 * sizeof(unsigned long int))
#define Q_PUSH(low, high) ((void) ((top->lo = (low)), (top->hi = (high)), ++top))
#define Q_POP(low, high)  ((void) (--top, (low = top->lo), (high = top->hi)))
#define Q_STACK_NOT_EMPTY (stack < top)                


/* Order size using quicksort.  This implementation incorporates
   four optimizations discussed in Sedgewick:

   1. Non-recursive, using an explicit stack of pointer that store the 
      next array partition to sort.  To save time, this maximum amount 
      of space required to store an array of MAX_INT is allocated on the 
      stack.  Assuming a 32-bit integer, this needs only 32 * 
      sizeof(qsort_stack_node) == 136 bits.  Pretty cheap, actually.

   2. Chose the pivot element using a median-of-three decision tree.
      This reduces the probability of selecting a bad pivot value and 
      eliminates certain extraneous comparisons.

   3. Only quicksorts TOTAL_ELEMS / Q_MAX_THRESH partitions, leaving
      insertion sort to order the Q_MAX_THRESH items within each partition.  
      This is a big win, since insertion sort is faster for small, mostly
      sorted array segements.

   4. The larger of the two sub-partitions is always pushed onto the
      stack first, with the algorithm then concentrating on the
      smaller partition.  This *guarantees* no more than log (n)
      stack size is needed (actually O(1) in this case)!  */

void qsort(void *pbase, size_t total_elems, size_t size, int(*cmp) (const void *keyval, const void *datum))
{
  register char *base_ptr = (char *) pbase;
  char *pivot_buffer;
  const size_t max_thresh = Q_MAX_THRESH * size;

  /* Allocating SIZE bytes for a pivot buffer facilitates a better
     algorithm below since we can do comparisons directly on the pivot. */

  pivot_buffer = (char *) Heap$Malloc (Pvs(heap), size);

  if (total_elems == 0) {
    /* Avoid lossage with unsigned arithmetic below.  */
    FREE (pivot_buffer);
    return;
  }

  if (total_elems > Q_MAX_THRESH)
    {
      char *lo = base_ptr;
      char *hi = &lo[size * (total_elems - 1)];
      /* Largest size needed for 32-bit int!!! */
      qsort_stack_node stack[Q_STACK_SIZE];
      qsort_stack_node *top = stack + 1;

      while (Q_STACK_NOT_EMPTY)
        {
          char *left_ptr;
          char *right_ptr;

          char *pivot = pivot_buffer;

          /* Select median value from among LO, MID, and HI. Rearrange
             LO and HI so the three values are sorted. This lowers the 
             probability of picking a pathological pivot value and 
             skips a comparison for both the LEFT_PTR and RIGHT_PTR. */

          char *mid = lo + size * ((hi - lo) / size >> 1);

          if ((*cmp)((void*) mid, (void*) lo) < 0)
            Q_SWAP(mid, lo, size);
          if ((*cmp)((void*) hi, (void*) mid) < 0)
            Q_SWAP(mid, hi, size);
          else 
            goto jump_over;
          if ((*cmp)((void*) mid, (void*) lo) < 0)
            Q_SWAP(mid, lo, size);
        jump_over:;
          memcpy(pivot, mid, size);
          pivot = pivot_buffer;

          left_ptr  = lo + size;
          right_ptr = hi - size; 

          /* Here's the famous ``collapse the walls'' section of quicksort.  
             Gotta like those tight inner loops!  They are the main reason 
             that this algorithm runs much faster than others. */
          do 
            {
              while ((*cmp)((void*) left_ptr, (void*) pivot) < 0)
                left_ptr += size;

              while ((*cmp)((void*) pivot, (void*) right_ptr) < 0)
                right_ptr -= size;

              if (left_ptr < right_ptr) 
                {
                  Q_SWAP(left_ptr, right_ptr, size);
                  left_ptr += size;
                  right_ptr -= size;
                }
              else if (left_ptr == right_ptr) 
                {
                  left_ptr += size;
                  right_ptr -= size;
                  break;
                }
            } 
          while (left_ptr <= right_ptr);

          /* Set up pointers for next iteration.  First determine whether
             left and right partitions are below the threshold size.  If so, 
             ignore one or both.  Otherwise, push the larger partition's
             bounds on the stack and continue sorting the smaller one. */

          if ((size_t) (right_ptr - lo) <= max_thresh)
            {
              if ((size_t) (hi - left_ptr) <= max_thresh)
                /* Ignore both small partitions. */
                Q_POP(lo, hi); 
              else
                /* Ignore small left partition. */  
                lo = left_ptr;
            }
          else if ((size_t) (hi - left_ptr) <= max_thresh)
            /* Ignore small right partition. */
            hi = right_ptr;
          else if ((right_ptr - lo) > (hi - left_ptr))
            {                   
              /* Push larger left partition indices. */
              Q_PUSH(lo, right_ptr);
              lo = left_ptr;
            }
          else
            {                   
              /* Push larger right partition indices. */
              Q_PUSH(left_ptr, hi);
              hi = right_ptr;
            }
        }
    }

  /* Once the BASE_PTR array is partially sorted by quicksort the rest
     is completely sorted using insertion sort, since this is efficient 
     for partitions below Q_MAX_THRESH size. BASE_PTR points to the beginning 
     of the array to sort, and END_PTR points at the very last element in
     the array (*not* one beyond it!). */

  /* MIN is #defined in nemesis.h. */

  {
    char *const end_ptr = &base_ptr[size * (total_elems - 1)];
    char *tmp_ptr = base_ptr;
    char *thresh = MIN(end_ptr, base_ptr + max_thresh);
    register char *run_ptr;

    /* Find smallest element in first threshold and place it at the
       array's beginning.  This is the smallest array element,
       and the operation speeds up insertion sort's inner loop. */

    for (run_ptr = tmp_ptr + size; run_ptr <= thresh; run_ptr += size)
      if ((*cmp)((void*) run_ptr, (void*) tmp_ptr) < 0)
        tmp_ptr = run_ptr;

    if (tmp_ptr != base_ptr)
      Q_SWAP(tmp_ptr, base_ptr, size);

    /* Insertion sort, running from left-hand-side up to right-hand-side.  */

    run_ptr = base_ptr + size;
    while ((run_ptr += size) <= end_ptr)
      {
        tmp_ptr = run_ptr - size;
        while ((*cmp)((void*) run_ptr, (void*) tmp_ptr) < 0)
          tmp_ptr -= size;

        tmp_ptr += size;
        if (tmp_ptr != run_ptr)
          {
            char *trav;

            trav = run_ptr + size;
            while (--trav >= run_ptr)
              {
                char c = *trav;
                char *hi, *lo;

                for (hi = lo = trav; (lo -= size) >= tmp_ptr; hi = lo)
                  *hi = *lo;
                *hi = c;
              }
          }
      }
  }

  FREE (pivot_buffer);

}

#endif


};
#endif



#ifdef TEST_LIST

#include <stdio.h>

int compare(const void *e1, const void *e2)
{
	char *s1 = *(char **)e1;
	char *s2 = *(char **)e2;
//	printf("s1= %s\n", s1);
//	printf("s2= %s\n", s2);
	int c = strcmp(s1, s2);
//	printf("comparison of %s,%s returns %d\n", s1, s2, c);
	return c;
//		return strcmp((const char *)e1, (const char *)e2);
}


void dump(List &list)
{
	int l = list.length();
	printf("  length() = %d\n", l);
	for(int i = 0; i < l; i++)
	{
		char *s = (char *)list.get(i);
		printf("  %d: %s\n", i, s);
	}
}

int main(int argc, char **argv)
{
	List list;

	for(int i = 1; i < argc; i++)
	{
		list.append(argv[i]);
	}

	printf("raw list:\n");
	dump(list);

	list.sort(compare);
	printf("sorted list:\n");
	dump(list);


	char *foo = "foo";
	char *bar = "bar";
	char *baz = "baz";

	printf("inserting foo at head, bar in middle and baz at end:\n");
	list.insert(foo, 0);
	list.insert(bar, list.length() / 2);
	list.insert(baz, list.length());
	dump(list);

	printf("sorting again:\n");
	list.sort(compare);
	dump(list);

	printf("removing foo (tests locate()):\n");
	list.rem(foo);
	dump(list);

	printf("removing 0th element:\n");
	list.rem(0);
	dump(list);

	printf("removing all:\n");
	list.remAll();
	dump(list);

}
#endif




#endif	// com_sleepless_list_cpp
