
// Copyright 2002
// Sleepless Software Inc.
// All Rights Reserved

#ifndef com_sleepless_net_sslsocket_h
#define com_sleepless_net_sslsocket_h


#include <openssl/err.h>
#include <openssl/rand.h>
#include <openssl/ssl.h>

#include "../socket.cpp"



static SSL_CTX *ctx = 0;
static bool sslGate = false;


static void sslTrouble(char *s)
{
	fprintf(stderr,"%s\n", s);
    ERR_print_errors_fp(stderr);
}


static void sslInit(const char *randfile, const char *cafile)
{
	if(!sslGate)
	{
		// Init ssl library
		SSL_library_init();
		SSL_load_error_strings();

		// Create context
		ctx = SSL_CTX_new(SSLv23_method());
		if(ctx == NULL)
		{
			sslTrouble("can't create ctx\n");
			return;
		}

		// Load trusted certificate authorities
		if(cafile)
		{
			int vl = SSL_CTX_load_verify_locations(ctx, cafile, 0);
			if(!vl)
				sslTrouble("can't load CA file\n");
			SSL_CTX_set_verify_depth(ctx, 1);
		}

		// Load randomness
		if(randfile)
		{
			int lr = RAND_load_file(randfile, 1024 * 1024);
			if(!lr)
				sslTrouble("can't load randomness\n");
		}

		// Don't permit anything to happ if called twice
		sslGate = true;
	}
}


struct SSLSocket : public Socket
{
	SSL	*ssl;


	int available()
	{
		if(ssl)
			return SSL_pending(ssl);
		return 0;
	}


	int write(const char *buf, int len)
	{
		if(isConnected())
		{
			int r = SSL_write(ssl, buf, len);
			if(r < 0)
				sslTrouble("SSL write error");
			return r;
		}
		return -1;
	}


	/*int write(const char *str)
	{
		return write(str, strlen(str));
	}


	int write(int n)
	{
		char buf[64];
		sprintf(buf, "%d", n);
		return write(buf);
	}


	int write(float n)
	{
		char buf[64];
		sprintf(buf, "%f", n);
		return write(buf);
	}
	*/


	int read(char *buf, int len)
	{
		if(isConnected())
		{
			int r = 0;
			int n;
			while(r < len)
			{
				n = SSL_read(ssl, buf + r, len - r);
				if(n == -1)
					return -1;
				if(n == 0)
					break;
				r += n;
			}
			return r;
		}
		return -1;
	}


	int readLine(char *buf, int len)
	{
		if(isConnected())
		{
			int n = 0;
			int r;
			for( ; n < (len - 1); n++)
			{
				r = read(&(buf[n]), 1);
				if(r != 1)
				{
					buf[n] = 0;	// Terminate the string.
					break;
				}

				if(buf[n] == 13)		// Discard CR's 
				{
					n--;
					continue;
				}
			
				if(buf[n] == 10)		// LF marks the end of the line
				{
					n++;
					buf[n] = 0;	// Terminate the string.
					break;
				}

			}
			return n;
		}
		return -1;
	}


	void sslTrouble(char *s)
	{
		fprintf(stderr,"%s\n", s);
		ERR_print_errors_fp(stderr);
	}


	bool cInit(const char *cafile, const char *randfile, const char *host)
	{
	
		ssl = SSL_new(ctx);
		if(ssl == NULL)
		{
			sslTrouble("can't make new ssl context\n");
			return false;
		}

		BIO *sbio = BIO_new_socket(sock, BIO_NOCLOSE);
		SSL_set_bio(ssl, sbio, sbio);

		if(SSL_connect(ssl) <= 0)
		{
			sslTrouble("ssl can't connect\n");
			return false;
		}


		// Check certificate chain.
		if(SSL_get_verify_result(ssl) != X509_V_OK)
		{
			sslTrouble("certificate check chain fails\n");
			return false;
		}


		// Check the common name
		X509 *peer = SSL_get_peer_certificate(ssl);
		char peer_CN[256];
		X509_NAME_get_text_by_NID(X509_get_subject_name(peer), NID_commonName, peer_CN, 256);
		if(strcasecmp(peer_CN, host))
		{
			sslTrouble("common name doesn't match\n");
			return false;
		}

		return true;
	}


	SSLSocket(const char *host, int port, const char *randfile, const char *cafile) : Socket(host, port)
	{
		if(sock != INVALID_SOCKET)
		{
			sslInit(randfile, cafile);
			if(!cInit(cafile, randfile, host))
			{
				socketClose(sock);
				sock = INVALID_SOCKET;
			}
		}
	}

	bool sInit()
	{
		ssl = SSL_new(ctx);
		BIO *sbio = BIO_new_socket(sock, BIO_NOCLOSE);
		SSL_set_bio(ssl, sbio, sbio);
		if(SSL_accept(ssl) <= 0)
		{
			sslTrouble("SSL accept error");
			return false;
		}
		return true;
	}

	SSLSocket(SOCKET s, unsigned long addr, const char *randfile, const char *cafile) : Socket(s, addr)
	{
		if(sock != INVALID_SOCKET)
		{
			sslInit(randfile, cafile);
			if(!sInit())
			{
				socketClose(sock);
				sock = INVALID_SOCKET;
			}
		}
	}


	~SSLSocket()
	{
		if(ssl != NULL)
		{
			SSL_shutdown(ssl);
			SSL_free(ssl);
		}

		if(sock != INVALID_SOCKET)
		{
			socketClose(sock);
			sock = INVALID_SOCKET;
		}
	}

};



#ifdef SSLSOCKET_TEST

#include <stdio.h>
#include <stdlib.h>


int main(int argc, char **argv)
{
	if(argc < 5)
	{
		printf("Usage: %s host port randfile cafile\n", argv[0]);
		return 1;
	}

	char *host = argv[1];
	int port = atoi(argv[2]);
	char *randfile = argv[3];
	char *cafile = argv[4];


	SSLSocket *sock = new SSLSocket(host, port, randfile, cafile);
	if(!sock->isConnected())
	{
		printf("#### Can't connect to %s on port %d\n", host, port);
		return 1;
	}

	printf("Connected to %s on port %d\n", host, port);

	char line[256];

	fgets(line, sizeof(line), stdin);
	int r = sock->writeStr(line);
	if(r == -1)
	{
		printf("#### sock->writeStr() returned -1\n");
		exit(1);
	}
	fgets(line, sizeof(line), stdin);
	r = sock->writeStr(line);
	if(r == -1)
	{
		printf("#### sock->writeStr() returned -1\n");
		exit(1);
	}

	if(r == 0)
	{
		printf("EOF\n");
		exit(1);
	}


	while(true)
	{

		int r = sock->readLine(line, sizeof(line));

		if(r == -1)
		{
			printf("#### sock->read() returned -1\n");
			break; 
		}

		if(r == 0)
		{
			printf("EOF\n");
			break;
		}

		printf("%s", line);
	}

	delete sock;

	printf("closed\n");

	return 0;
}


#endif


#endif // com_sleepless_net_sslsocket_h
