#ifndef com_sleepless_sha_cpp
#define com_sleepless_sha_cpp


#include <stdio.h>
#include <string.h>



struct SHA
{
	typedef struct sha_digest { unsigned long A, B, C, D, E; } sha_digest;

	typedef unsigned char u8;
	typedef unsigned long u32;
#ifdef WIN32
	typedef __int64 u64;
#else
	typedef unsigned long long u64;
#endif


	void sha(const void *message, long size, sha_digest *out_dig)
	{
		u64 message_size = (u64)size * 8;
		u32 digA = 0x67452301;
		u32 digB = 0xEFCDAB89;
		u32 digC = 0x98BADCFE;
		u32 digD = 0x10325476;
		u32 digE = 0xC3D2E1F0;

		/* Loop through each 512-bit block of the message.  */
		while (1)
		{
			u32 A = digA;
			u32 B = digB;
			u32 C = digC;
			u32 D = digD;
			u32 E = digE;
			u32 W[80];
			u32 t;
			u32 nA;

			/* Find the contents of this block,
			   in the first 16 words of W[].  */
			if (size >= 512/8)
			{
				/* Supply the complete block
				   straight from the message.  */
				memcpy(W, message, 512/8);
			}
			else if (size >= (512-64)/8)
			{
				/* This is the last block of the message,
				   but there's not enough room for at least one pad bit
				   and the u64 holding the message length.
				   Just place the necessary padding in this block,
				   and there will be one final block afterwards.  */
				memcpy(W, message, size);
				((u8*)W)[size] = 0x80;
				memset((u8*)W+size+1, 0, 512/8-(size+1));
			}
			else if (size >= 0)
			{
				/* This is the last block of the message,
				   and there _is_ room for the padding and length.  */
				if (size > 0)
					memcpy(W, message, size);
				((u8*)W)[size] = 0x80;
				memset((u8*)W+size+1, 0, (512-64)/8-(size+1));
				*(u64*)(W+(512-64)/32) = message_size;
			}
			else if (size >= -64/8)
			{
				/* This is a trailer block,
				   holding no part of the message,
				   only padding and the message size.  */
				memset(W, 0, (512-64)/8);
				*(u64*)(W+(512-64)/32) = message_size;
			}
			else
			{
				/* We're done.  */
				break;
			}
			size -= 512/8;

			/* Generate the remaining four
			   of the first 20 values of W_t.  */
			for (t = 16; t < 20; t++)
				W[t] = W[t-16] ^ W[t-14] ^ W[t-8] ^ W[t-3];

			/* Run the digest buffer through the rounds.  */
			for (t = 0; t < 20; t++)
			{
				nA = ((A << 5) | (A >> 27))
					+ ((B & C) | (~B & D))
					+ E
					+ W[t]
					+ 0x5A827999;
				E = D;
				D = C;
				C = ((B << 30) | (B >> 2));
				B = A;
				A = nA;
			}
			for (; t < 40; t++)
			{
				W[t] = W[t-16] ^ W[t-14] ^ W[t-8] ^ W[t-3];
				nA = ((A << 5) | (A >> 27))
					+ (B ^ C ^ D)
					+ E
					+ W[t]
					+ 0x6ED9EBA1;
				E = D;
				D = C;
				C = ((B << 30) | (B >> 2));
				B = A;
				A = nA;
			}
			for (; t < 60; t++)
			{
				W[t] = W[t-16] ^ W[t-14] ^ W[t-8] ^ W[t-3];
				nA = ((A << 5) | (A >> 27))
					+ ((B & C) | (B & D) | (C & D))
					+ E
					+ W[t]
					+ 0x8F1BBCDC;
				E = D;
				D = C;
				C = ((B << 30) | (B >> 2));
				B = A;
				A = nA;
			}
			for (; t < 80; t++)
			{
				W[t] = W[t-16] ^ W[t-14] ^ W[t-8] ^ W[t-3];
				nA = ((A << 5) | (A >> 27))
					+ (B ^ C ^ D)
					+ E
					+ W[t]
					+ 0xCA62C1D6;
				E = D;
				D = C;
				C = ((B << 30) | (B >> 2));
				B = A;
				A = nA;
			}

			/* Add the final digest buffer
			   back into the overall digest.  */
			digA += A;
			digB += B;
			digC += C;
			digD += D;
			digE += E;
		}

		/* Return the final message digest.  */
		out_dig->A = digA;
		out_dig->B = digB;
		out_dig->C = digC;
		out_dig->D = digD;
		out_dig->E = digE;
	}

	// 'buf' must be at least (5 * 8) + 1 bytes long so that a null terminated
	// string something like "ffffffff00000000aaaaaaaa22222222" can be
	// written into it.
	void sha(const void *message, long size, char *buf)
	{
		sha_digest od;
		sha(message, size, &od);
		sprintf(buf, "%8.8lx%8.8lx%8.8lx%8.8lx%8.8lx", od.A, od.B, od.C, od.D, od.E);
	}


};

struct SHADigest
{
	char digest[(5 * 8) + 1];

	SHADigest(const char *data, unsigned long datalen)
	{
		SHA sha;
		sha.sha(data, datalen, digest);
	}

	const char *getBytes()
	{
		return digest;
	}

};



#ifdef SHA_TEST

#include "file.cpp"

int main(int argc, char **argv)
{

	if(argc != 2)
	{
		printf("Usage: sha filename|string\n");
		exit(0);
	}

	File f(argv[1]);
	if(f.isFile())
	{
		SHADigest digest((const char *)f.getBytes(), f.getSize());
		puts(digest.getBytes());
	}
	else
	{
		SHADigest digest(argv[1], strlen(argv[1]));
		puts(digest.getBytes());
	}

	exit(0);
}

#endif


#endif		// com_sleepless_sha_cpp
