#ifndef com_sleepless_misc_random
#define com_sleepless_misc_random



/* This random number generation code was found at
 * http://www.math.keio.ac.jp/~matumoto/emt.html
 * Hopefully it will work well for me. 
 * It claims to be 2 to 4 times faster than rand(), but
 * I find that on my linux 7.0 box, on a PII 400, it runs
 * a bit slower than rand().   100 million iterations in 23
 * seconds versus 21 seconds for rand().
 * Still, that's pretty fast I think.
 * And it looks like it might generate a better distribution.
 *
 * j.h.
 */

/* This library is free software; you can redistribute it and/or   */
/* modify it under the terms of the GNU Library General Public     */
/* License as published by the Free Software Foundation; either    */
/* version 2 of the License, or (at your option) any later         */
/* version.                                                        */
/* This library is distributed in the hope that it will be useful, */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of  */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.            */
/* See the GNU Library General Public License for more details.    */
/* You should have received a copy of the GNU Library General      */
/* Public License along with this library; if not, write to the    */
/* Free Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA   */ 
/* 02111-1307  USA                                                 */

/* Copyright (C) 1997, 1999 Makoto Matsumoto and Takuji Nishimura. */
/* Any feedback is very welcome. For any question, comments,       */
/* see http://www.math.keio.ac.jp/matumoto/emt.html or email       */
/* matumoto@math.keio.ac.jp                                        */

/* REFERENCE                                                       */
/* M. Matsumoto and T. Nishimura,                                  */
/* "Mersenne Twister: A 623-Dimensionally Equidistributed Uniform  */
/* Pseudo-Random Number Generator",                                */
/* ACM Transactions on Modeling and Computer Simulation,           */
/* Vol. 8, No. 1, January 1998, pp 3--30.                          */
#include <stdio.h>
#include <stdlib.h>

#ifdef __palmos__
#define RAND_MAX 2147483647
#endif


struct Random
{

/* Period parameters */  
//#define N 624
#define RND_N 624
//#define M 397
#define RND_M 397
#define MATRIX_A 0x9908b0df   /* constant vector a */
#define UPPER_MASK 0x80000000 /* most significant w-r bits */
#define LOWER_MASK 0x7fffffff /* least significant r bits */

/* Tempering parameters */   
#define TEMPERING_MASK_B 0x9d2c5680
#define TEMPERING_MASK_C 0xefc60000
#define TEMPERING_SHIFT_U(y)  (y >> 11)
#define TEMPERING_SHIFT_S(y)  (y << 7)
#define TEMPERING_SHIFT_T(y)  (y << 15)
#define TEMPERING_SHIFT_L(y)  (y >> 18)

	unsigned long mt[RND_N]; /* the array for the state vector  */
	int mti;


	Random()
	{
		mti = RND_N+1; /* mti==RND_N+1 means mt[RND_N] is not initialized */
		setSeed(13);
	}


	/* Initializing the array with a seed */
	/* SEED SHOULD NOT BE 0! */
	void setSeed(unsigned long seed)
	{
		int i;

		for (i=0;i<RND_N;i++) {
			 mt[i] = seed & 0xffff0000;
			 seed = 69069 * seed + 1;
			 mt[i] |= (seed & 0xffff0000) >> 16;
			 seed = 69069 * seed + 1;
		}
		mti = RND_N;
	}

	/* Initialization by "setSeed()" is an example. Theoretically,      */
	/* there are 2^19937-1 possible states as an intial state.           */
	/* This function allows to choose any of 2^19937-1 ones.             */
	/* Essential bits in "seed_array[]" is following 19937 bits:         */
	/*  (seed_array[0]&UPPER_MASK), seed_array[1], ..., seed_array[RND_N-1]. */
	/* (seed_array[0]&LOWER_MASK) is discarded.                          */ 
	/* Theoretically,                                                    */
	/*  (seed_array[0]&UPPER_MASK), seed_array[1], ..., seed_array[RND_N-1]  */
	/* can take any values except all zeros.                             */
	void lsgenrand(
			unsigned long seed_array[] /* the length of seed_array[] must be at least RND_N */
			)
	{
		int i;

		for (i=0;i<RND_N;i++) 
		  mt[i] = seed_array[i];
		mti=RND_N;
	}


	double nextDouble()
	{
		unsigned long y;
		//const static unsigned long mag01[2]={0x0, MATRIX_A};
		unsigned long mag01[2]; mag01[0]=0x0; mag01[1]=MATRIX_A;
		/* mag01[x] = x * MATRIX_A  for x=0,1 */

		if (mti >= RND_N) { /* generate RND_N words at one time */
			int kk;

			if (mti == RND_N+1)   /* if setSeed() has not been called, */
				setSeed(4357); /* a default initial seed is used   */

			for (kk=0;kk<RND_N-RND_M;kk++) {
				y = (mt[kk]&UPPER_MASK)|(mt[kk+1]&LOWER_MASK);
				mt[kk] = mt[kk+RND_M] ^ (y >> 1) ^ mag01[y & 0x1];
			}
			for (;kk<RND_N-1;kk++) {
				y = (mt[kk]&UPPER_MASK)|(mt[kk+1]&LOWER_MASK);
				mt[kk] = mt[kk+(RND_M-RND_N)] ^ (y >> 1) ^ mag01[y & 0x1];
			}
			y = (mt[RND_N-1]&UPPER_MASK)|(mt[0]&LOWER_MASK);
			mt[RND_N-1] = mt[RND_M-1] ^ (y >> 1) ^ mag01[y & 0x1];

			mti = 0;
		}
	  
		y = mt[mti++];
		y ^= TEMPERING_SHIFT_U(y);
		y ^= TEMPERING_SHIFT_S(y) & TEMPERING_MASK_B;
		y ^= TEMPERING_SHIFT_T(y) & TEMPERING_MASK_C;
		y ^= TEMPERING_SHIFT_L(y);

		return ( (double)y * 2.3283064370807974e-10 ); /* reals */
	}


	unsigned long nextLong()
	{
		unsigned long y;
		//const static unsigned long mag01[2]={0x0, MATRIX_A};
		unsigned long mag01[2]; mag01[0]=0x0; mag01[1]=MATRIX_A;
		/* mag01[x] = x * MATRIX_A  for x=0,1 */

		if (mti >= RND_N) { /* generate RND_N words at one time */
			int kk;

			if (mti == RND_N+1)   /* if setSeed() has not been called, */
				setSeed(4357); /* a default initial seed is used   */

			for (kk=0;kk<RND_N-RND_M;kk++) {
				y = (mt[kk]&UPPER_MASK)|(mt[kk+1]&LOWER_MASK);
				mt[kk] = mt[kk+RND_M] ^ (y >> 1) ^ mag01[y & 0x1];
			}
			for (;kk<RND_N-1;kk++) {
				y = (mt[kk]&UPPER_MASK)|(mt[kk+1]&LOWER_MASK);
				mt[kk] = mt[kk+(RND_M-RND_N)] ^ (y >> 1) ^ mag01[y & 0x1];
			}
			y = (mt[RND_N-1]&UPPER_MASK)|(mt[0]&LOWER_MASK);
			mt[RND_N-1] = mt[RND_M-1] ^ (y >> 1) ^ mag01[y & 0x1];

			mti = 0;
		}
	  
		y = mt[mti++];
		y ^= TEMPERING_SHIFT_U(y);
		y ^= TEMPERING_SHIFT_S(y) & TEMPERING_MASK_B;
		y ^= TEMPERING_SHIFT_T(y) & TEMPERING_MASK_C;
		y ^= TEMPERING_SHIFT_L(y);

		return y;
	}


	// Return a random int like the standard rand() does.
	int rand()
	{
		unsigned long l = nextLong();
		return (int)(l % RAND_MAX);
	}


	/*int roll(int n)
	{
		return rand() % n;
	}*/

	/* Roll an n-sided die.
	For example, if a 4 is passed in, 0, 1, 2 or 3 will be returned.*/
	unsigned long roll(unsigned long n)
	{
		return nextLong() % n;
	}

};


#ifdef TEST
#include <stdlib.h>
#include <time.h>
int main(int argc, char *argv)
{ 
	Random random;
    long i;

	printf("Printing first 1000 random numbers ...\n");
    random.setSeed(4357);
    for (i=0; i<1000; i++)
	{
        printf("%10.8f ", random.nextDouble());
        if (i%5==4)
			printf("\n");
    }

	printf("Testing speed against C-lib version of rand()\n");

	unsigned long start;
	unsigned long stop;
	unsigned long t;
	int r;

	srand(12345);
	start = time(0);
    for (i=0; i < 100000000; i++)
	{
		r = rand();
	}
	stop = time(0);

	t = stop - start;		// seconds
	printf("C-lib rand() took %ld seconds\n", stop - start);


	random.setSeed(12345);
	start = time(0);
    for (i=0; i < 100000000; i++)
	{
		r = (int)random.nextLong();
	}
	stop = time(0);

	t = stop - start;		// seconds
	printf("MT PRNG took %ld seconds\n", stop - start);

}
#endif





#if 0
// Uncertain if this version really works well or not.
struct OldRandom
{
    long multiplier;
    long addend;
    long mask;
    long seed;

    OldRandom(long seed) {
		multiplier = (long)0x5DEECE66D;
		addend = (long)0xB;
		mask = ((long)1 << 48) - 1;
        setSeed(13);
    }

    void setSeed(long s) {
        seed = (s ^ multiplier) & mask;
    }

    int next(int bits) {
        long nextseed = (seed * multiplier + addend) & mask;
        seed = nextseed;
		long n = nextseed >> (48 - bits);
		if(n < 0)
			n *= -1;
        return n;
    }

    long nextLong() {
        // it's okay that the bottom word remains signed.
        return ((long)(next(32)) << 32) + next(32);
    }

};
#endif



#endif		// com_sleepless_misc_random
