#ifndef com_sleepless_serversocket_cpp
#define com_sleepless_serversocket_cpp

/* Copyright 2004  Sleepless Software Inc.  All Rights Reserved */

#include "socket.cpp"

struct ServerSocket
{

	SERVERSOCKET ss;


	/* Returns true if the server socket is set up and
	ready to accept new connections.*/
	virtual bool isListening()
	{
		return ss != INVALID_SOCKET;
	}


	/* Tries to create an open and bound server socket
	with the provided backlog parameter.
	Call the method isListening() to determine if
	the socket is ready to receive connections. */
	ServerSocket(int port, int backlog)
	{
		ss = serverSocketOpen(port, backlog);
	}

	ServerSocket(const char *ipstr, int port, int backlog)
	{
		ss = serverSocketOpen(ipstr, port, backlog);
	}

	virtual ~ServerSocket()
	{
		serverSocketClose(ss);
		ss = INVALID_SOCKET;
	}


	/* Returns a ptr to a new Socket object for a new incoming
	connection on the server socket or null if some error occured.
	NOTE: This call is going to block until a new
	connection comes in.  It just calls accept(). */
	virtual Socket *accept()
	{
		struct sockaddr_in sa;
		memset(&(sa), 0, sizeof(sa)); 
		socklen_t salen = sizeof(sa);
		SOCKET s = serverSocketAccept(ss, &sa, &salen);
		if(s == INVALID_SOCKET)
			return NULL;
		unsigned long addr = ntohl(sa.sin_addr.s_addr);
		return new Socket(s, addr);
	}


	/* Same as accept(), but will give up after 'maxsecs' of waiting.
	Returns null if the time expired before getting a connection, so there
	is no way to know if the accept failed, or the time expired. */
	virtual Socket *accept(int maxsecs)
	{
		struct sockaddr_in sa;
		memset(&(sa), 0, sizeof(sa)); 
		socklen_t salen = sizeof(sa);
		SOCKET s = serverSocketAccept(ss, &sa, &salen, maxsecs);
		if(s == INVALID_SOCKET)
			return NULL;
		unsigned long addr = ntohl(sa.sin_addr.s_addr);
		return new Socket(s, addr);
	}


	/* Returns true if a connection is pending ... a call to accept()
	should immediately return a new socket. */
	virtual int pending()
	{
		return serverSocketPending(ss);
	}

};



#ifdef SERVER_SOCKET_TEST

#include <stdio.h>
#include <stdlib.h>

//#include "./socket.cpp"

int main(int argc, char **argv)
{
	int port = 9000;

	if(argc >= 2)
		port = atoi(argv[1]);

	ServerSocket ss(port, 1);
	if(!ss.isListening())
	{
		printf("#### Can't create ServerSocket on port %d\n", port);
		return 1;
	}

	printf("Listening on port %d\n", port);

	Socket *sock = ss.accept();
	if(!sock)
	{
		printf("#### Bad accept()\n");
		return 1;
	}

	printf("Got connection ...\n");

	char line[256];
	while(true)
	{
		int r = sock->readLine(line, sizeof(line));

		if(r == -1)
		{
			printf("#### readLine() returned -1\n");
			return 1;
		}

		if(r == 0)
		{
			printf("Got EOF\n");
			break;
		}

		printf("read: %s", line);

		r = sock->writeStr(line);

		if(r == -1)
		{
			printf("#### sock->writeStr() returned -1\n");
			break; 
		}

		if(r == 0)
		{
			printf("#### sock->writeStr() returned 0\n");
			break;
		}

	}

	delete sock;

	printf("closed\n");

	return 0;
}

#endif



#endif // com_sleepless_serversocket_cpp
