#ifndef sha1_cpp
#define sha1_cpp

/*
	Based on the sha1 code by Michael D. Leonhard found at

		http://tamale.net/sha1/

	which is licensed thusly

		http://www.opensource.org/licenses/mit-license.php
*/


//#include <string.h>
//#include <stdlib.h>

#include "assert.cpp"

class SHA1
{
	typedef unsigned int Uint32;

	private:
		Uint32			H0, H1, H2, H3, H4;
		unsigned char	bytes[64];
		int				unprocessedBytes;
		Uint32			size;
		unsigned char	_digest[20];
		char			_ascii[41];

		void process()
		{
			ASSERT( unprocessedBytes == 64 );

			int t;
			Uint32 a, b, c, d, e, K, f, W[80];

			// starting values
			a = H0;
			b = H1;
			c = H2;
			d = H3;
			e = H4;

			// copy and expand the message block
			for( t = 0; t < 16; t++ ) W[t] = (bytes[t*4] << 24)
											+(bytes[t*4 + 1] << 16)
											+(bytes[t*4 + 2] << 8)
											+ bytes[t*4 + 3];
			for(; t< 80; t++ ) W[t] = lrot( W[t-3]^W[t-8]^W[t-14]^W[t-16], 1 );
			
			/* main loop */
			Uint32 temp;
			for( t = 0; t < 80; t++ )
			{
				if( t < 20 ) {
					K = 0x5a827999;
					f = (b & c) | ((b ^ 0xFFFFFFFF) & d); //TODO: try using ~
				} else if( t < 40 ) {
					K = 0x6ed9eba1;
					f = b ^ c ^ d;
				} else if( t < 60 ) {
					K = 0x8f1bbcdc;
					f = (b & c) | (b & d) | (c & d);
				} else {
					K = 0xca62c1d6;
					f = b ^ c ^ d;
				}
				temp = lrot(a,5) + f + e + W[t] + K;
				e = d;
				d = c;
				c = lrot(b,30);
				b = a;
				a = temp;
			}

			/* add variables */
			H0 += a;
			H1 += b;
			H2 += c;
			H3 += d;
			H4 += e;
			
			/* all bytes have been processed */
			unprocessedBytes = 0;
		}

	public:

		SHA1()
		{
			// make sure that the data type is the right size
			ASSERT( sizeof( Uint32 ) * 5 == 20 );
			
			// initialize
			H0 = 0x67452301;
			H1 = 0xefcdab89;
			H2 = 0x98badcfe;
			H3 = 0x10325476;
			H4 = 0xc3d2e1f0;
			unprocessedBytes = 0;
			size = 0;
		}

		~SHA1()
		{
			// erase data
			H0 = H1 = H2 = H3 = H4 = 0;
			for( int c = 0; c < 64; c++ )
				bytes[c] = 0;
			unprocessedBytes = size = 0;
		}


		void add( const char* data, int num )
		{
			if(num < 1)
				return;
			ASSERT( data );
			//ASSERT( num > 0 );
			// add these bytes to the running total
			size += num;
			// repeat until all data is processed
			while( num > 0 )
			{
				// number of bytes required to complete block
				int needed = 64 - unprocessedBytes;
				ASSERT( needed > 0 );
				// number of bytes to copy (use smaller of two)
				int toCopy = (num < needed) ? num : needed;
				// Copy the bytes
				memcpy( bytes + unprocessedBytes, data, toCopy );
				// Bytes have been copied
				num -= toCopy;
				data += toCopy;
				unprocessedBytes += toCopy;
				
				// there is a full block
				if( unprocessedBytes == 64 ) process();
			}
		}

		const unsigned char *digest()
		{
			// save the message size
			Uint32 totalBitsL = size << 3;
			Uint32 totalBitsH = size >> 29;
			// add 0x80 to the message
			add( "\x80", 1 );
			
			unsigned char footer[64] = {
				0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
				0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
				0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
				0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
			// block has no room for 8-byte filesize, so finish it
			if( unprocessedBytes > 56 )
				add( (char*)footer, 64 - unprocessedBytes);
			ASSERT( unprocessedBytes <= 56 );
			// how many zeros do we need
			int neededZeros = 56 - unprocessedBytes;
			// store file size (in bits) in big-endian format
			storeBigEndianUint32( footer + neededZeros    , totalBitsH );
			storeBigEndianUint32( footer + neededZeros + 4, totalBitsL );
			// finish the final block
			add( (char*)footer, neededZeros + 8 );
			// copy the digest bytes
			storeBigEndianUint32( _digest, H0 );
			storeBigEndianUint32( _digest + 4, H1 );
			storeBigEndianUint32( _digest + 8, H2 );
			storeBigEndianUint32( _digest + 12, H3 );
			storeBigEndianUint32( _digest + 16, H4 );
			// return the digest
			return _digest;
		}

		const char *asciiDigest()
		{
			const char *hexdigits = "0123456789abcdef";
			const unsigned char *d = digest();
			char *a = _ascii;
			for(int i = 0; i < 20; i++)
			{
				char c = *d++;
				*a++ = hexdigits[(c & 0xf0) >> 4];
				*a++ = hexdigits[c & 0x0f];
			}
			_ascii[40] = 0;
			return _ascii;
		}

		// circular left bit rotation.  MSB wraps around to LSB
		static Uint32 lrot( Uint32 x, int bits )
		{
			return (x<<bits) | (x>>(32 - bits));
		}

		// Save a 32-bit unsigned integer to memory, in big-endian order
		static void storeBigEndianUint32(unsigned char* byte, Uint32 num)
		{
			ASSERT( byte );
			byte[0] = (unsigned char)(num>>24);
			byte[1] = (unsigned char)(num>>16);
			byte[2] = (unsigned char)(num>>8);
			byte[3] = (unsigned char)num;
		}

		// print out memory in hexadecimal
		/*
		static void hexPrinter(unsigned char* c, int l)
		{
			ASSERT( c );
			ASSERT( l > 0 );
			while( l > 0 )
			{
				printf( "%02x", *c );
				l--;
				c++;
			}
		}
		*/
};


#ifdef SHA1_TEST

#include <string.h>
#include <stdlib.h>

int main(int argc, char *argv[])
{
	SHA1 sha1;
	if(argc == 2)
	{
		sha1.add(argv[1], strlen(argv[1]));
		puts(sha1.asciiDigest());
	}
	else
	{
		char buf[10000];
		while(!feof(stdin))
		{
			int n = fread(buf, 1, sizeof(buf), stdin);
			sha1.add( buf, n );
		}
		puts(sha1.asciiDigest());
	}
	exit(0);
}

#endif


#endif	// sha1_cpp
