#ifndef com_sleepless_sound_cpp
#define com_sleepless_sound_cpp

/* Copyright 2003 Sleepless Software Inc. All Rights Reserved */



//#define sfx_samp_rate		11025
//#define sfx_num_chans		1		// (mono)
//#define sfx_samp_bits		16		
//#define sfx_samp_align		((sfx_samp_bits / 8) * sfx_num_chans)
#ifndef sfx_num_slots
#define sfx_num_slots		8		// max simultaneously playing snds
#endif


#include "assert.cpp"
#include "file.cpp"


#ifdef WIN32

#define DIRECTDRAW_VERSION 0x0700
#include <dsound.h>

#include "mutex.cpp"
#include "windows.cpp"

static LPDIRECTSOUND sfx_dsnd = NULL;
static IDirectSoundBuffer *sfx_slots[sfx_num_slots];
static Mutex sfx_mutex;
static int sfx_active_slots = 0;


static void sfx_dsnd_init()
{
	int r;
	r = DirectSoundCreate(NULL, &sfx_dsnd, NULL);
	DASSERT(r == DS_OK);
	HWND w = Windows::createBlandWindow("sleepy-sound", 10, 10, DefWindowProc);
	DASSERT(w != NULL);
	r = sfx_dsnd->SetCooperativeLevel(w, DSSCL_NORMAL);
	DASSERT(r == DS_OK);
}

#endif



struct Sound
{

#ifdef WIN32
	IDirectSoundBuffer *dsb;
#endif


	void init(unsigned char *data, int rate, int bits, int len, int chans)
	{
		int align = (bits / 8) * chans;
		len = (len / align) * align;
		DASSERT(len > 0);
		DASSERT(data);

#ifdef WIN32

		dsb = 0;

		if(!sfx_dsnd)
			sfx_dsnd_init();

		// create directsound buffer
		WAVEFORMATEX wfmt;
		wfmt.wFormatTag = WAVE_FORMAT_PCM; 
		wfmt.nChannels = chans;
		wfmt.nSamplesPerSec = rate; 
		wfmt.wBitsPerSample = bits;
		wfmt.nBlockAlign = align;
		wfmt.nAvgBytesPerSec = rate * align;
		wfmt.cbSize = 0;
		DSBUFFERDESC bd;
		bd.dwSize = sizeof(bd); 
		bd.dwFlags = DSBCAPS_CTRLVOLUME|DSBCAPS_CTRLFREQUENCY|DSBCAPS_GLOBALFOCUS;
		bd.dwBufferBytes = len;
		bd.dwReserved = 0;
		bd.lpwfxFormat = &wfmt; 
		int r = sfx_dsnd->CreateSoundBuffer(&bd, &dsb, NULL);
		DASSERT(r == DS_OK);

		// copy sample data into directsound buffer
		DWORD sz1, sz2;
		void *p1, *p2;
		r = dsb->Lock(0, len, &p1, &sz1, &p2, &sz2, 0);
		DASSERT(r == DS_OK);
		memcpy(p1, data, sz1);
		if(p2)
			memcpy(p2, data + sz1, sz2);
		dsb->Unlock(p1, sz1, p2, sz2);
#endif
	}


	// File provided MUST contain plain, raw samples in this format:
	//		Mono
	//		16-bit, signed samples
	// 		11025 samples/sec
	Sound(const char *file)
	{
		File f(file);
		int len = f.diskSize(); //  / sfx_samp_align) * sfx_samp_align;
		unsigned char *data = f.getBytes();

		init(data, 11025, 16, len, 1);
	}


	Sound(unsigned char *data, int rate, int bits, int len, int chans)
	{
		init(data, rate, bits, len, chans);
	}

	
	~Sound()
	{
#ifdef WIN32
		if(dsb)
			dsb->Release();
#endif
	}


	// volume = 1.0 for full volume
	// volume = 0.0 for attenuated to silence
	// volume = 0.5 is about half loudness
	void play(float volume)
	{

		if(volume <= 0)
			return;
	
		if(volume > 1)
			volume = 1;

#ifdef WIN32
		if(!dsb)
			return;
		
		sfx_mutex.lock();

		sfx_active_slots = 0;
		for(int i = 0; i < sfx_num_slots; i++)
		{
			IDirectSoundBuffer *b = sfx_slots[i];
			
			// reap slots that have finished as we go through the array
			if(b)
			{
				unsigned long status;
  				int r = b->GetStatus(&status);
				DASSERT(r == DS_OK);
				if(!(status & DSBSTATUS_PLAYING))
				{
					// sample has finished playing
					b->Release();
					b = sfx_slots[i] = NULL;
				}
			}

			// now see if this slot is free to re-use
			if(b == NULL)
			{
				IDirectSoundBuffer *ndsb;
				int r = sfx_dsnd->DuplicateSoundBuffer(dsb, &ndsb);
				if(r == DS_OK)
				{
					ndsb->SetVolume((long)((4000 * volume) - 4000));
					ndsb->Play(0, 0, 0);
					sfx_slots[i] = ndsb;
					break;
				}
				else
				{
					ASSERT(0);
				}
			}

			if(sfx_slots[i])
				sfx_active_slots++;
		}

		sfx_mutex.unlock();
#endif
	}


	void play()
	{
		play(1);
	}

};


#endif	// com_sleepless_sound_cpp
