#ifndef com_sleepless_swreg_h
#define com_sleepless_swreg_h

// Copyright 1998-2004
// Sleepless Software Inc.
// All Rights Reserved

/*

How the reg. key/code scheme works.

1. Game trial expires:

	The # of runs is stored in the "timestamp" of a file associated
	with the game such as "data/bg.bmp".  So the game stores how 
	many times it's been run by "setting" the timestamp of this
	data file to be a specific # of seconds "different" from the
	time stamp on the game executable.  It's offset a little
	to avoid the problem of the dates being very similar
	freshly after installation.

2. Software Registration

	- Customer pays for registration at kagi/paypal site
		which is accessed from the game's home page and
		from the game itself via a link/button on the 
		nag/trial-over screen.

	- Registrant's data (name, address, transaction # etc) is
		sent from kagi/paypal to joe via email when the
		transaction is processed.

	- The info is stored into the DB and the "reg data" 
		is generated from it, which looks something like this:

			c2d738e0f34fe5ed6712fcbaff571732e93764bd
			iw
			2004-04-20 10:21:43
			Marcus
			Naylor
			62 Gresham Road
			Hillingdon
			Uxbridge
			Middlesex
			UB10 OHU
			United Kingdom

			marcus.naylor@blueyonder.co.uk
			CHDFMZ53GM2E

		The reg data is a blob of text with a bunch of lines.
		Every line except for the first line, is
		the relevant data about a purchase and purchaser,
		one data point per line.
		The first line is a secure hash of the rest of the lines.
		A secure hash (SHA) is similar in concept to a checksum.
		It's unique and easy to generate from the data, but it's
		essentially impossible to derive the original data
		from the hash.

	- A "reg code" is generated from the "reg data"
		The reg code is the 3rd character, and every
		4th character following.
		So for a hash like:
			d01fcc2e1c1a0c44cb426dd8aaafd58721246910
		the reg code is:
			12144da821

	- The reg code is sent to the purchaser as an "unlock code".
		The reg code rather than the has itself is sent simply
		because that's a shitload of characters to type just
		exactly so, and because the game's unlock code field
		doesn't support CTRL-V for pasting.

3. Unlocking the software

	- The user enters the reg code into the game's unlock dialog.

	- The unlock code is submitted to the sleepless site via an
		HTTP transaction.
		Note that the user has to be online to unlock the game.

	- The server side looks up the
		database record of the purchase using the regcode and
		returns the reg data blob to the game.

	- The game stores reg data in a local file called "rgdt.txt"

	- The game restarts itself.

	- On startup, the game checks for "rgdt.txt" and
		reads it in.  It generates a hash from all of the
		data minus the first line.  If the generated hash
		matches the hash on the first line, then the game
		assumes the game is registered.

4. Notes

	As we both know, if someone wants to crack the game, they can
	and there really isn't much we can do to prevent it.

	Also, this system does NOT prevent a user from copying their
	reg data to a friends game dir and thereby making that game
	"registered".  However, if they do so, then the name and
	address of the ORIGINAL purchaser MUST be carried along with
	the illegal version (because of the hash).  The idea is that
	when someone purchases an unlock code, it is made clear
	to them that their unlock code contains personally identifying
	information with it. Hopefully that will cause them to think
	twice about copying the reg data.  Also, since we don't actually
	tell anyone that you CAN copy the reg data, there isn't 
	much reason for to even suspect that they could.

	FURTHER NOTE: At the moment, there is actually an additional
	hole in the security ... the reg data for any game will
	unlock any other game using this system.  For example, if
	you register Pocoman, you can copy your poco rgdt.txt to
	the nuts dir and unlock it that way.  This isn't hard to fix,
	I just haven't gotten round to it yet. (-:

*/


#include <ctype.h>

#include "misc.cpp"
#include "sha.cpp"
#include "file.cpp"


struct SWReg
{
	// return a copy of 'in' with all white space chars removed
	static char *crunched(const char *in)
	{
		int len = strlen(in) + 1;
		char *out = new char[len];
		if(out)
		{
			int o = 0;
			int i = 0;
			while(i < len)
			{
				if(!isspace(in[i]))
				{
					out[o] = in[i];
					o++;
				}
				i++;
			}
		}
		return out;
	}

	static int verify(const char *path)
	{
		int r = 1;	// failure

		File file(path);
		const unsigned char *fbytes = file.getBytes();
		if(fbytes)
		{
			char *fhash = strdup((const char *)fbytes);
			if(fhash)
			{
				char *data = strchr(fhash, 10);
				if(data)
				{
					*data = 0;
					Str::trim(fhash);
					data++;
					char *crdata = crunched(data);
					if(crdata)
					{
						SHADigest digest(crdata, strlen(crdata));
						const char *hash = digest.getBytes();
						if(strcmp(fhash, hash) == 0)
							r = 0;	// OK
						delete crdata;
					}
				}
			}
		}

		return r;
	}

	// The 'data' argument must point to a null terminated string.
	// This function won't work for binary data.
	// Returns 0 on success
	static int regcode(const char *path)
	{
		int r = 1;	// failure

		File file(path);
		const char *data = (const char *)file.getBytes();
		if(data)
		{
			char *crdata = crunched(data);
			if(crdata)
			{
				SHADigest digest(crdata, strlen(crdata));
				const char *hash = digest.getBytes();

				int l = strlen(hash);
				for(int i = 2; i < l; i += 4)
					fprintf(stdout, "%c", hash[i]);
				fprintf(stdout, "\n");

				r = 0;
				delete crdata;
			}
		}

		return r;
	}


	// The 'data' argument must point to a null terminated string.
	// This function won't work for binary data.
	// Returns 0 on success
	static int create(const char *path)
	{
		int r = 1;	// failure

		File file(path);
		const char *data = (const char *)file.getBytes();
		if(data)
		{
			char *crdata = crunched(data);
			if(crdata)
			{
				SHADigest digest(crdata, strlen(crdata));
				const char *hash = digest.getBytes();
				fprintf(stdout, "%s%c%s", hash, 10, data);
				r = 0;
				delete crdata;
			}
		}

		return r;
	}


};


#ifdef CLI_SWREG

#include <stdio.h>

int main(int argc, char **argv)
{
	if(argc != 3)
	{
		printf(
"Usage: swreg [create|regcode|verify] file\n\
   create : 'file' contains reg data, hashed reg data written to stdout\n\
   regcode: 'file' contains reg data, reg code written to stdout\n\
   verify : 'file' contains hashed reg data, exit code 0 if data is kosher\n\
");
		exit(1);
	}


	const char *cmd = argv[1];
	const char *file = argv[2];


	int r = 1;
	
	if(tolower(cmd[0]) == 'c')
	{
		r = SWReg::create(file);
	}
	else
	if(tolower(cmd[0]) == 'r')
	{
		r = SWReg::regcode(file);
	}
	else
	if(tolower(cmd[0]) == 'v')
	{
		r = SWReg::verify(file);
	}
	else
	{
		fprintf(stderr, "Unrecognized command: %s\n", cmd);
	}

	exit(r);
	return r;
}


#endif



#endif	// com_sleepless_swreg_h
