#ifndef com_sleepless_tagarray_h
#define com_sleepless_tagarray_h

// Copyright 1998-2004  Sleepless Software Inc.  All Rights Reserved

/*	Similar to an Array, but you reference the pointers on it with a
	text 'tag' rather than by index number.

	Also similar to TagList except that:
	
		- this object is for storing ptrs to anything, not just strings
			so you have to cast your ptrs when taking them off.

		- Also, this object doesn't have the support for reading/writing
			to disk files.

		- you can access things by index with this object, but not 
			with TagList

	*/

//#include "misc.cpp"
#include "list.cpp"

struct TagArray
{
	List tags;
	List vals;


	/* Returns the number of pointers currently in this tagarray. */
	int length()
	{
		return tags.length();
	}


	/* Puts a pointer "somewhere" on the shelf along with the given tag.
		If there is already a pointer on the shelf with the same tag, the old
		pointer is removed and returned and the new pointer and tag will
		replace it.  If there wasn't already a pointer with the tag on the
		list, null is returned. */
	void *put(const char *tag, void *val)
	{
		char *t = strdup(tag);
		void *oldval = rem(t);
		tags.append((void *)t);
		vals.append(val);
		return oldval;
	}


	/* Returns the pointer on the shelf with the given tag without removing
		it, or NULL if the tag isn't found.
		Tag comparison is case _insensitive_. */
	void *get(const char *tag)
	{
		int l = length();
		for(int i = 0; i < l; i++)
		{
			if(strcasecmp((const char *)tags.get(i), tag) == 0)
				return vals.get(i);
		}
		return 0;
	}


	/* Returns the pointer at index number 'i' without removing it, or NULL
		if there isn't a pointer at that index or the index is invalid. */
	void *get(int i)
	{
		return vals.get(i);		// Array (inside List) does range check
	}


	/* Removes and returns a pointer given it's tag, or null if not found */
	void *rem(const char *tag)
	{
		int l = length();
		for(int i = 0; i < l; i++)
		{
			if(strcmp((const char *)tags.get(i), tag) == 0)
			{
				char *t = (char *)tags.rem(i);
				if(t)
					free(t);
				return vals.rem(i);
			}
		}
		return 0;
	}


	/* Removes and returns the pointer at index 'i' or null if index invalid */
	void *rem(int i)
	{
		char *t = (char *)tags.rem(i);	// if this fails and returns null ...
		if(t)
			free(t);
		return vals.rem(i);		// so should this
	}


	/* Removes and returns any pointer or null if the array is empty. */
	void *rem()
	{
		return rem(0);
	}



	/* Returns the tag for the pointer at the given index or null if there
		isn't a pointer at that index or the index is invalid. */
	const char *getTag(int i)
	{
		return (const char *)tags.get(i);
	}


	/* Moves all of the pointers from another TagArray to this one.
		The other will be empty afterward. */
	void put(TagArray *ots)
	{
		int l = ots->length();
		for(int i = 0; i < l; i++)
			put(ots->getTag(i), ots->get(i));
		ots->clear();
	}

	void clear()
	{
		int l = length();
		for(int i = 0; i < l; i++)
			rem(0);
	}

};




#ifdef TAGARRAY_TEST


TagArray ta;
TagArray ta2;

void dump()
{
	int l = ta.length();
	printf(" Mary -- length() returns %d\n", l);
	for(int i = 0; i < l; i++)
	{
		printf("  %d: tag = '%s'  -  val = '%s'\n", i, ta.getTag(i), ta.get(i));
	}
	printf("\n");
}

void dump2()
{
	int l = ta2.length();
	printf(" Jane -- length() returns %d\n", l);
	for(int i = 0; i < l; i++)
	{
		printf("  %d: tag = '%s'  -  val = '%s'\n", i, ta2.getTag(i), ta2.get(i));
	}
	printf("\n");
}


int main(int argc, char **argv)
{
	char *name = "name";
	char *addr = "addr";
	char *phone = "phone";
	char *email = "email";
	
	printf("initially:\n");
	dump();

	printf("adding Mary's name, addr and phone\n");
	ta.put(name, (void *)"Mary");
	ta.put(addr, (void *)"123 Elm St");
	ta.put(phone, (void *)"555-1212");
	dump();

	printf("mary's addr is %s\n", ta.get(addr));
	printf("mary's name is %s\n", ta.get(name));
	printf("mary's phone # is %s\n", ta.get(phone));
	printf("mary's email addr is %s\n", ta.get(email));

	printf("removing addr ... %s\n", ta.rem(addr));
	dump();

	printf("removing 0th item ... %s\n", ta.rem(0));
	dump();

	printf("removing any item ... %s\n", ta.rem());
	dump();

	printf("adding Jane's name, addr and phone\n");
	ta2.put(name, (void *)"Jane");
	ta2.put(addr, (void *)"99 Oak Dr");
	ta2.put(phone, (void *)"pennsylvania 6-5000");
	dump2();

	printf("copying Jane's info from other tagarray ...\n");
	ta.put(&ta2);
	dump();
	dump2();
}


#endif




#endif	// com_sleepless_tagarray_h
