#ifndef com_sleepless_unigate_cpp
#define com_sleepless_unigate_cpp

// Copyright 2010 Sleepless Software Incorporated All Rights Reserved

/*
	The constuctor arg is a path to a file in the filesystem.  That file-system path
	given in the constructor can only be "locked" (see lock()) by one unigate,
	regardless of which program or process attempts the lock.  Unigate therefore,
	can be used something like a system-wide mutex.  It may even work across networked
	filesystems in some cases, but I haven't tried that.

	The lock is released if:
		* the unlock() method is called 
		* the lock(const char *g) method is called (unlocks, then relocks with g)
		* the unigate object is deleted.
		* the process that created the object exits

	Example:

		void foo()
		{
			Unigate ug("file/system/path");
			if(ug.lock())
			{
				// Do stuff that only one program should ever be doing at the same time
			}
			else
			{
				// Someone else locked the unigate before I could. I'll just do something else.
			}
			// ug is destroyed and lock released when foo() returns.
		}


*/

#ifdef WIN32
//#include <sys/types.h>
//#include <sys/stat.h>
#include <io.h>
#else
#include <unistd.h>
#endif
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>


struct Unigate
{
	char *gate;
	int fd;

	void unlock()
	{
		if(gate)
		{
			if(fd >= 0)
			{
				close(fd);
				fd = -1;
			}
		}
	}

	bool lock()
	{
		if(!gate)
			return false;
#ifdef WIN32
		fd = open(gate, O_CREAT|O_WRONLY, 0666);
#else
		fd = open(gate, O_NONBLOCK|O_CREAT|O_EXLOCK|O_WRONLY, 0666);
#endif
		return fd >= 0;
	}

	bool lock(const char *g)
	{
		unlock();
		if(gate)
			free(gate);
		gate = strdup(g);
		return lock();
	}

	~Unigate()
	{
		unlock();
		if(gate)
			free(gate);
	}

	Unigate(const char *g)
	{
		gate = strdup(g);
		fd = -1;
	}

	Unigate()
	{
		gate = 0;
		fd = -1;
	}

};



#ifdef TEST_UNIGATE

int main()
{
	Unigate ug("/tmp/foo");
	if(ug.lock())
	{
		printf("locked acquired\n");
		sleep(5);
		ug.unlock();
		printf("locked released\n");
		sleep(5);
		return 0;
	}
	
	printf("failed\n");

	return 0;
}

#endif


#endif // com_sleepless_unigate_cpp
